#import "MCGraphicView.h"
#import "MCEvent.h"
#include <pthread.h>
#include "mlvalues.h"

extern code_t start_code;
extern MCGraphicView* gv;
int interprete(code_t);

void* interprete1(void* p)
{
    interprete((code_t)p);
    [NSApp terminate:nil];
    return NULL;
}

@implementation MCGraphicView
@synthesize graphicWindow;
@synthesize image;
@synthesize queue;
@synthesize isHidden;
@synthesize data;
@synthesize bitmap;
@synthesize graphicContextInitialized;
NSString* serverName;
BOOL activate = YES;

- (void)applicationDidFinishLaunching:(NSApplication *)application
{
    [self.window makeKeyAndOrderFront:nil];
    NSConnection* connection = [NSConnection defaultConnection];
    [connection enableMultipleThreads];
    connection.rootObject = gv;
    int count = 1;
    while ([connection registerName:(serverName = [NSString stringWithFormat:@"MacCamlServer%d", count])] == NO && count < 10000)
    {
        count++;
    }
    if (count >= 10000) NSLog(@"couldn't register MacCaml graphic server");
    pthread_t th;
    pthread_create(&th, NULL, &interprete1, start_code);
}

- (void)applicationDidUpdate:(NSNotification *)notification {
    if (activate) {[NSApp activateIgnoringOtherApps:YES]; activate = NO;}
}

- (id)initWithFrame:(NSRect)frame andTitle:(NSString*)title
{
    graphicContextInitialized = NO;
    if (self = [super initWithFrame:frame])
    {
        graphicWindow = [[NSWindow alloc]
                         initWithContentRect: frame
                         styleMask: NSTitledWindowMask
                         | NSClosableWindowMask
                         | NSMiniaturizableWindowMask
                         | NSTexturedBackgroundWindowMask
                         backing: NSBackingStoreBuffered
                         defer: NO];
        NSRect r = [NSScreen.mainScreen visibleFrame];
        r.origin.x += frame.origin.x;
        r.origin.y += r.size.height - frame.origin.y;
        [graphicWindow setFrameTopLeftPoint:r.origin];
        graphicWindow.title = title;
        graphicWindow.delegate = (id<NSWindowDelegate>)self;
        graphicWindow.releasedWhenClosed = NO;
        graphicWindow.initialFirstResponder = self;
        image = [[NSImage alloc] initWithSize:frame.size];
        queue = [[NSMutableArray alloc] initWithCapacity:100];
        bitmap = nil;
        data = nil;
        isHidden = NO;
    }
    return self;
}
- (void) setServerFrame:(NSRect)rect title:(NSString*)t;
{
    graphicWindow.contentView = self;
    graphicWindow.contentView.appearance = [NSAppearance appearanceNamed:NSAppearanceNameAqua];
    image.size = rect.size;
    self.window.contentSize = rect.size;
    self.window.title = t;
    [self.window makeKeyAndOrderFront:nil];
    isHidden = NO;
    NSRect r = NSScreen.mainScreen.visibleFrame;
    r.origin.x += rect.origin.x;
    r.origin.y += r.size.height - rect.origin.y;
    [self.window setFrameTopLeftPoint:r.origin];
}

- (void) getImage:(NSData*) im
{
    if (im)
    {
        NSRect rect = self.frame;
        int width = rect.size.width;
        int height = rect.size.height;
        int byteCount = width * height * 4;
        if (bitmap != nil){[image removeRepresentation:bitmap];}
        free(data);
        data = (unsigned char*)malloc(byteCount);
        bitmap = [[NSBitmapImageRep alloc]
                  initWithBitmapDataPlanes:&data
                  pixelsWide:width
                  pixelsHigh:height
                  bitsPerSample:8
                  samplesPerPixel:4 hasAlpha:YES
                  isPlanar:NO
                  colorSpaceName:NSCalibratedRGBColorSpace
                  bytesPerRow:(4 * width)
                  bitsPerPixel:32];
        [im getBytes: data length: byteCount];
        [image addRepresentation:bitmap];
        if (self.window.isMiniaturized) {self.window.miniwindowImage = image;}
        self.needsDisplay = YES;
    }
} 

- (void)drawRect:(NSRect)rect
{
    if (graphicContextInitialized == NO)
    {
        NSGraphicsContext* gc = NSGraphicsContext.currentContext;
        if (gc) {
            [NSColor.whiteColor set];
            NSRectFill(rect);
            gc.shouldAntialias = NO;
            graphicContextInitialized = YES;
        }
    }
    if (data) [image drawInRect:rect fromRect:rect operation:NSCompositeCopy fraction:1.0];
}

- (void) hideWindow;
{
    isHidden = YES;
    [self.window close];
}

-(BOOL)acceptsFirstResponder
{
    return YES;
}

-(BOOL) isOpaque
{
    return YES;
}

- (MCEvent*) getNextEvent
{
    if (queue.count == 0) {return nil;}
    MCEvent* mce = [queue objectAtIndex:0];
    MCEvent* mce1 = [[MCEvent alloc]init];
    mce1.mask = mce.mask;
    mce1.mouse_x = mce.mouse_x;
    mce1.mouse_y = mce.mouse_y;
    mce1.key = mce.key;
    [queue removeObjectAtIndex:0];
    return mce1;
}

- (void) windowDidDeminiaturize:(NSNotification *)aNotification
{
    [self.window makeKeyAndOrderFront:nil];
}

- (void)windowWillClose:(NSNotification *)notification
{
    if (isHidden == YES) return;
    MCEvent* mce = [[MCEvent alloc]init];
    mce.mask = WINDOWCLOSED;
    mce.mouse_x = 0;
    mce.mouse_y = 0;
    mce.key = 0;
    [queue addObject:mce];
    graphicContextInitialized = NO;
}

- (void) mouseDown:(NSEvent*) e
{
    MCEvent* mce = [[MCEvent alloc]init];
    NSPoint pt = [self convertPoint:e.locationInWindow fromView:nil];
    mce.mask = MOUSEDOWN;
    mce.mouse_x = pt.x;
    mce.mouse_y = pt.y;
    mce.key = 0;
    [queue addObject:mce];
}

- (void) mouseUp:(NSEvent*) e
{
    MCEvent* mce = [[MCEvent alloc]init];
    NSPoint pt = [self convertPoint:e.locationInWindow fromView:nil];
    mce.mask = MOUSEUP;
    mce.mouse_x = pt.x;
    mce.mouse_y = pt.y;
    mce.key = 0;
    [queue addObject:mce];
}

- (void)mouseDragged:(NSEvent*)e
{
    MCEvent* mce = [[MCEvent alloc]init];
    NSPoint pt = [self convertPoint:e.locationInWindow fromView:nil];
    mce.mask = MOUSEMOVE;
    mce.mouse_x = pt.x;
    mce.mouse_y = pt.y;
    mce.key = 0;
    [queue addObject:mce];
}

- (void) keyDown:(NSEvent*) e
{
    MCEvent* mce = [[MCEvent alloc]init];
    NSPoint pt = [self convertPoint:self.window.mouseLocationOutsideOfEventStream fromView:nil];
    mce.mask = KEYDOWN;
    mce.mouse_x = pt.x;
    mce.mouse_y = pt.y;
    mce.key = [e.characters characterAtIndex:0];
    [queue addObject:mce];
}

- (void) dealloc
{
    graphicWindow.delegate = nil;
    graphicWindow.releasedWhenClosed = YES;
    graphicWindow = nil;
    free(data);
}

@end

