(*-------------------------------------------------------------------------------------------*)
(* TYPES *)
(*-------------------------------------------------------------------------------------------*)

type mouvement1 =
	{mutable mv1: (int vect * int vect vect) list}
;;

type context =
	{mutable matrice: int vect vect}
;;

type ops = OPS of (unit -> unit) * (unit -> unit) * (unit -> unit);;

type couleur = ORANGE | VERT | BLANC | ROUGE | BLEU | JAUNE | GRIS;;
type face = ANTERIEURE | DROITE | HAUTE | POSTERIEURE | GAUCHE | BASSE;;

type bouton = {titre: string;
		orx: int; ory: int;
		largeur: int; hauteur: int;
		mutable couleur: couleur;
		mutable action: unit -> unit};;

(* type hybride pour saisir un mouvement *)
type info = {mutable videG: bool; mutable videD: bool; mutable ct: int vect; mutable mat: int vect vect};;


(*-------------------------------------------------------------------------------------------*)
(* DIVERS *)
(*-------------------------------------------------------------------------------------------*)

(* quelques fonctions auxiliaires sur les matrices en dimension 3 *)

let vect v = if vect_length v = 3 then (v.(0), v.(1), v.(2))
	else failwith "vect"
;;

let identity a = let m = make_matrix 3 3 0 in
		for i = 0 to 2 do
			m.(i).(i) <- a
		done;
		m
;;

let id = identity 1 and idm = identity (- 1);;

(* produit du vecteur ligne entier v par la matrice entière m *)
let prefix /:/ v m =
	let w j = let t = ref 0 in for k = 0 to vect_length v - 1 do
				t := !t + m.(k).(j) * v.(k) done;
			!t in
		[|w 0; w 1; w 2|]
;;

(* produit matriciel *)
let prefix /./ m m1 = map_vect (fun v -> v /:/ m1) m;;

(* somme matricielle *)
let prefix /+/ m1 m2 =
	let m = make_matrix 3 3 0 in
		for i = 0 to 2 do
			for j = 0 to 2 do
				m.(i).(j) <- m1.(i).(j) + m2.(i).(j)
			done;
		done;
		m
;;

(* produit de la colonne v par la ligne w *)
let prefix /::/ v w =
	let m = make_matrix 3 3 0 in
		for i = 0 to 2 do
			for j = 0 to 2 do
				m.(i).(j) <- v.(i) * w.(j)
			done;
		done;
		m
;;

(* matrice diagonale *)
let diag a b c = [|[|a; 0; 0|]; [|0; b; 0|]; [|0; 0; c|]|];;

(* transposée de la matrice m  qui en est aussi l'inverse : *)
(* quand m est orthogonale *)
let transpose m =
	let m1 = make_matrix 3 3 0 in
		for i = 0 to 2 do
			for j = 0 to 2 do
				m1.(j).(i) <- m.(i).(j)
			done;
		done;
		m1
;;

(* produit scalaire *)
let prefix /|/ v w = v.(0) * w.(0) + v.(1) * w.(1) + v.(2) * w.(2);;

(* matrices des rotations d'un quart de tour autour des axes : *)
(* (opèrent à droite sur les lignes) *)

(* sens des aiguilles d'une montre *)
let rot v = match list_of_vect v with
		| [1; 0; 0] -> [|[|1; 0; 0|]; [|0; 0; - 1|]; [|0; 1; 0|]|]
		| [0; 1; 0] -> [|[|0; 0; 1|]; [|0; 1; 0|]; [|- 1; 0; 0|]|]
		| [0; 0; 1] -> [|[|0; - 1; 0|]; [|1; 0; 0|]; [|0; 0; 1|]|]
		| [- 1; 0; 0] -> [|[|1; 0; 0|]; [|0; 0; 1|]; [|0; - 1; 0|]|]
		| [0; - 1; 0] -> [|[|0; 0; - 1|]; [|0; 1; 0|]; [|1; 0; 0|]|]
		| [0; 0; - 1] -> [|[|0; 1; 0|]; [|- 1; 0; 0|]; [|0; 0; 1|]|]
		| _ -> failwith "rot"
;;

(* sens inverse des aiguilles d'une montre *)
let rot' v = match list_of_vect v with
		| [1; 0; 0] -> [|[|1; 0; 0|]; [|0; 0; 1|]; [|0; - 1; 0|]|]
		| [0; 1; 0] -> [|[|0; 0; - 1|]; [|0; 1; 0|]; [|1; 0; 0|]|]
		| [0; 0; 1] -> [|[|0; 1; 0|]; [|- 1; 0; 0|]; [|0; 0; 1|]|]
		| [- 1; 0; 0] -> [|[|1; 0; 0|]; [|0; 0; - 1|]; [|0; 1; 0|]|]
		| [0; - 1; 0] -> [|[|0; 0; 1|]; [|0; 1; 0|]; [|- 1; 0; 0|]|]
		| [0; 0; - 1] -> [|[|0; - 1; 0|]; [|1; 0; 0|]; [|0; 0; 1|]|]
		| _ -> failwith "rot'"
;;

(* liste dans l'ordre des éléments de l satisfaisant 'critère' *)
let rec select critere l = match l with
		t :: r -> let l1 = select critere r in if critere t then t :: l1 else l1
		| _ -> []
;;

(* liste des entiers de 0 à n - 1 *)
let liste n =
	let v = make_vect n 0 in
		for i = 0 to n - 1 do
			v.(i) <- i
		done;
		list_of_vect v
;;

(* permutation aléatoire des éléments d'une liste l *)
let random_list l =
	let n = list_length l and l1 = ref []
	in
		for i = 0 to n - 1 do
			l1 := (vect_of_list (subtract l !l1)).(random__int (n - i)) :: !l1
		done;
		!l1
;;

(* signature de la permutation p des éléments de la liste l *)
let sign l p =
	let n = list_length l and v = vect_of_list l
	and m = ref 1 in
		for i = 0 to n - 1 do
			for j = i + 1 to n - 1 do
				let a = v.(i) and b = v.(j) in
					if p a > p b && b > a || p b > p a && a > b then m := - !m;
			done;
		done;
		!m
;;

(* exécution d'une liste de mouvements *)
let rec exe l = match l with
		t :: r -> t (); exe r;
		| [] -> ()
;;

(* enregistrement sur disque d'un mouvement *)
let enregistrer mouvement file =
	let chan_out = open_out_bin file in
		output_value chan_out mouvement;
		close_out chan_out
;;

(* lecture sur disque d'un mouvement *)
let lire_mouvement file =
	try
		let chan_in = open_in_bin file
		in
			let mv1_saved =
				input_value chan_in
			in
				close_in chan_in;
				mv1_saved
	with sys__Sys_error s -> failwith s
;;


(*-------------------------------------------------------------------------------------------*)
(* INDICES *)
(*-------------------------------------------------------------------------------------------*)

(* indices des coins *)
let indices = let l = ref [] in
    for k = 1 downto - 1 do
      for j = 1 downto - 1 do
        for i = 1 downto - 1 do l := [|i; j; k|] :: !l
        done
      done
    done;
    select (fun x -> x.(0) * x.(1) * x.(2) <> 0) !l
;;


(*-------------------------------------------------------------------------------------------*)
(* GROUPE DES MOUVEMENTS *)
(*-------------------------------------------------------------------------------------------*)

(* groupe M des mouvements des minicubes *)

(* élément neutre de M *)
let e = map (fun x -> x, id) indices;;

(* conversion entre mouvement représenté par une fonction et mouvement *)
(* représenté par une liste : (int vect * int vect vect) list *)
let mv1_of_fun f =
  map (fun (x, y) -> (x, y /./ (f x))) e
;;
let fun_of_mv1 mv1 x =
  assoc x mv1
;;

let trier mv1 = sort__sort (fun x y -> fst x < fst y) mv1;;

(* loi interne *)
let prefix /*/ mv1 mv1' =
  let f = fun_of_mv1 mv1 and f' = fun_of_mv1 mv1'
  in
    let s t = t /:/ (f t)
    in trier (mv1_of_fun (fun x -> (f x) /./ (f' (s x))))
;;

(* inverse d'un élément *)
let inverse mv1 = map (fun (x, y) -> (x /:/ y, transpose y)) mv1;;

(* mouvements de Rubik élémentaires *)

(* rotations dans le sens des aiguilles d'une montre d'un quart de tour de la *)
(* face - tranche interne dans le cas du cube 4x4 - normale au vecteur sortant 'v' *)
let rub v = mv1_of_fun
  (fun x -> if (x /|/ v) = 1 then rot v else id)
;;

(* mouvement inverse du précédent *)
let rub' v = inverse (rub v);;


(*-------------------------------------------------------------------------------------------*)
(* GROUPE DU CUBE *)
(*-------------------------------------------------------------------------------------------*)

let groupe_du_cube =
  [
    [|[|1; 0; 0|]; [|0; 1; 0|]; [|0; 0; 1|]|];
    [|[|1; 0; 0|]; [|0; - 1; 0|]; [|0; 0; - 1|]|];
    [|[|- 1; 0; 0|]; [|0; 1; 0|]; [|0; 0; - 1|]|];
    [|[|- 1; 0; 0|]; [|0; - 1; 0|]; [|0; 0; 1|]|];
    [|[|1; 0; 0|]; [|0; 0; - 1|]; [|0; 1; 0|]|];
    [|[|0; 0; 1|]; [|0; 1; 0|]; [|- 1; 0; 0|]|];
    [|[|0; - 1; 0|]; [|1; 0; 0|]; [|0; 0; 1|]|];
    [|[|1; 0; 0|]; [|0; 0; 1|]; [|0; - 1; 0|]|];
    [|[|0; 0; - 1|]; [|0; 1; 0|]; [|1; 0; 0|]|];
    [|[|0; 1; 0|]; [|- 1; 0; 0|]; [|0; 0; 1|]|];
    [|[|0; - 1; 0|]; [|- 1; 0; 0|]; [|0; 0; - 1|]|];
    [|[|0; 1; 0|]; [|1; 0; 0|]; [|0; 0; - 1|]|];
    [|[|- 1; 0; 0|]; [|0; 0; - 1|]; [|0; - 1; 0|]|];
    [|[|0; 0; - 1|]; [|0; - 1; 0|]; [|- 1; 0; 0|]|];
    [|[|0; 0; 1|]; [|0; - 1; 0|]; [|1; 0; 0|]|];
    [|[|- 1; 0; 0|]; [|0; 0; 1|]; [|0; 1; 0|]|];
    [|[|0; 1; 0|]; [|0; 0; 1|]; [|1; 0; 0|]|];
    [|[|0; 0; - 1|]; [|- 1; 0; 0|]; [|0; 1; 0|]|];
    [|[|0; 0; 1|]; [|- 1; 0; 0|]; [|0; - 1; 0|]|];
    [|[|0; 1; 0|]; [|0; 0; - 1|]; [|- 1; 0; 0|]|];
    [|[|0; 0; - 1|]; [|1; 0; 0|]; [|0; - 1; 0|]|];
    [|[|0; - 1; 0|]; [|0; 0; - 1|]; [|1; 0; 0|]|];
    [|[|0; - 1; 0|]; [|0; 0; 1|]; [|- 1; 0; 0|]|];
    [|[|0; 0; 1|]; [|1; 0; 0|]; [|0; 1; 0|]|]
  ]
;;


(*-------------------------------------------------------------------------------------------*)
(* COULEURS *)
(*-------------------------------------------------------------------------------------------*)
(* couleur binaire de la  couleur c *)
let couleur_binaire_de_couleur c =
  match c with
    | ROUGE -> graphics__red (*rouge*)
    | ORANGE -> graphics__rgb 255 165 0 (* orange *)
    | BLEU -> graphics__rgb 0 150 225 (* bleu *)
    | VERT -> graphics__green (*vert *)
    | JAUNE -> graphics__yellow (*jaune *)
    | BLANC -> graphics__white (* blanc *)
    | GRIS -> graphics__rgb 100 100 100
;;

let couleur_de v =
  match vect v with
    | 1, 0, 0 -> ORANGE
    | - 1, 0, 0 -> ROUGE
    | 0, 1, 0 -> VERT
    | 0, - 1, 0 -> BLEU
    | 0, 0, 1 -> BLANC
    | 0, 0, - 1 -> JAUNE
    | _ -> GRIS
;;


let couleur_binaire_de v =
  couleur_binaire_de_couleur (couleur_de v)
;;


(*-------------------------------------------------------------------------------------------*)
(* GRAPHISME *)
(*-------------------------------------------------------------------------------------------*)

let ux = ref 10;;
let uy = ref 10;;
let ox = ref 153;;
let oy = ref 160;;

let proj x y z =
  let c = sqrt 6. /. 2. in
    (c *. (y -. x) /. sqrt 2., c *. (-. (x +. y) +. 2. *. z) /. sqrt 6.)
;;

let xx v pt3 =
  let (x, y, z) = vect (map_vect float_of_int pt3) in
    let (x1, y1, z1) =
      if v /|/ [|1; 1; 1|] = 1 then (x, y, z)
      else match vect v with
          | (_, 0, 0) -> (x -. 7., y, z)
          | (0, _, 0) -> (x, y -. 7., z)
          | _ -> (x, y, z -. 7.)
    in
      int_of_float (float_of_int !ox +. fst (proj x1 y1 z1) *. float_of_int !ux)
;;

let yy v pt3 =
  let (x, y, z) = vect (map_vect float_of_int pt3) in
    let (x1, y1, z1) =
      if v /|/ [|1; 1; 1|] = 1 then (x, y, z)
      else match vect v with
          | (_, 0, 0) -> (x -. 7., y, z)
          | (0, _, 0) -> (x, y -. 7., z)
          | _ -> (x, y, z -. 7.)
    in
      int_of_float (float_of_int !oy +. snd (proj x1 y1 z1) *. float_of_int !uy)
;;

(* la fonction 'drawPoly' est utilisée pour tracer le pourtour des projections *)
(* des faces des minicubes *)
let drawPoly poly =
  let (x, y) = poly.(0) in graphics__moveto x y;
    for i = 1 to vect_length poly - 1 do
      let (x, y) = poly.(i) in graphics__lineto x y
    done;
    let (x, y) = poly.(0) in graphics__lineto x y;
;;

(* la fonction 'draw' est utilisée pour dessiner la projection 'x' d'une face *)
(* de minicube en superposant le tracé du pourtour à la couleur de remplissage *)
let draw x =
  let a, b = x in
    graphics__set_color b;
    graphics__fill_poly a;
    graphics__set_color graphics__black;
    drawPoly a
;;

(* 'face v c' renvoie, si le minicube à l'emplacement d'indice 'c' a une face F *)
(* dans la face du Rubik's cube normale au vecteur sortant 'v', sous forme de vecteur *)
(* une liste circulaire des 4 sommets de F *)

let coeff = ref 1;;

let face v c =
  let e = v /|/ [|1; 1; 1|] in let w = [|e; e; e|] in
      let w1 = w /:/ rot v in
        let w2 = w1 /:/ rot v in
          let w3 = w2 /:/ rot v in
            let l = [w; w1; w2; w3] in
              let add m = for i = 0 to 2 do m.(i) <- m.(i) + !coeff * c.(i) done
              in
                do_list add l;
                vect_of_list l;
;;


(*-------------------------------------------------------------------------------------------*)
(* BOUTONS *)
(*-------------------------------------------------------------------------------------------*)
(* Gestion par boutons des mouvements globaux et des mouvements de Rubik *)

let bouton titre orx ory largeur hauteur couleur action =
  {titre = titre; orx = orx; ory = ory; hauteur = hauteur; largeur = largeur;
    couleur = couleur; action = action}
;;

let inverse_bouton b =
  graphics__set_color graphics__black;
  graphics__fill_rect b.orx b.ory b.largeur b.hauteur;
  let (x, y) = graphics__text_size b.titre in
    graphics__moveto (b.orx + (b.largeur - x) / 2) (b.ory + (b.hauteur - y) / 2);
    graphics__set_color graphics__white;
    graphics__draw_string b.titre;
    let p = make_vect 4 (0, 0) in
      p.(0) <- (b.orx, b.ory);
      p.(1) <- (b.orx + b.largeur, b.ory);
      p.(2) <- (b.orx + b.largeur, b.ory + b.hauteur);
      p.(3) <- (b.orx, b.ory + b.hauteur);
      graphics__set_color graphics__black;
      drawPoly p
;;

let dessine_bouton b =
  graphics__set_color (couleur_binaire_de_couleur (b.couleur));
  graphics__fill_rect b.orx b.ory b.largeur b.hauteur;
  let (x, y) = graphics__text_size b.titre in
    graphics__moveto (b.orx + (b.largeur - x) / 2) (b.ory + (b.hauteur - y) / 2);
    graphics__set_color graphics__black;
    graphics__draw_string b.titre;
    let p = make_vect 4 (0, 0) in
      p.(0) <- (b.orx, b.ory);
      p.(1) <- (b.orx + b.largeur, b.ory);
      p.(2) <- (b.orx + b.largeur, b.ory + b.hauteur);
      p.(3) <- (b.orx, b.ory + b.hauteur);
      graphics__set_color graphics__black;
      drawPoly p
;;

let set_action bouton action =
  bouton.action <- action
;;

let gestion_bouton bouton mousex mousey =
  if bouton.orx < mousex && mousex < bouton.orx + bouton.largeur
    && bouton.ory < mousey && mousey < bouton.ory + bouton.hauteur then (
      inverse_bouton bouton;
      let t1 = sys__time () in
        let t2 = ref t1 in
          while !t2 <. t1 +. 0.2 do
            t2 := sys__time ()
          done;
          dessine_bouton bouton;
          bouton.action ();
    );
;;


(*-------------------------------------------------------------------------------------------*)
(* SAISIE D'UN MOUVEMENT ou ÉTAT *)
(*-------------------------------------------------------------------------------------------*)

let mouvement = {mv1 = map (fun x -> (x, id)) indices};;(* élément neutre e du groupe des mouvements *)
let hauteur = 306;;
let largeur = 612;;

(*- saisie d'une configuration du cube 2x2 à partir des couleurs des faces visibles des minicubes -*)

(* 'faces' renvoie une liste de triplets : la première composante est un centre 'c', la deuxième composante *)
(* est un vecteur listant les 3 vecteurs unitaires sortants normaux aux faces visibles du minicube centré en 'c' *)
(* et la troisième est un vecteur dont la composante numéro i est un vecteur listant les 4 sommets de la face visible *)
(* normale au vecteur numéro i précédent *)

let faces c =
  let d = diag c.(0) c.(1) c.(2) in
    c, d, map_vect (fun v -> face v c) d
;;

let matrice_nulle = [|[|0; 0; 0|]; [|0; 0; 0|]; [|0; 0; 0|]|];;

(* construction du mouvement inverse par transferts de minicubes de gauche à droite à partir de la position neutre *)
(* ct est un centre de minicube et mat est l'inverse de la matrice de déplacement du minicube centré en ct *)
(* VideG = true signifie que le minicube centré en ct a été transféré à droite à l'emplacement centré en ct/:/mat *)
(* initialement tous les minicubes sont à gauche à leurs places *)
(* VideD = true signifie qu'à droite l'emplacement centré en ct est vide : au départ tous les emplacements à droite sont vides *)
let mvi = vect_of_list (map (fun x -> {videG = false; videD = true; ct = x; mat = matrice_nulle}) indices);;

(* affichage d'un minicube réorienté par la matrice de rotation mat *)
let affiche orx ory mat centre =
  let c, d, f = faces centre in
    for i = 0 to 2 do
      let v = d.(i) /:/ mat in
        draw ((map_vect (fun pt -> (xx v pt + orx, yy v pt + ory))
            (map_vect (fun x -> x /:/ mat) f.(i))),
          couleur_binaire_de d.(i))
    done
;;

(* affichage en gris d'un minicube *)
let griser orx ory centre =
  let c, d, f = faces centre in
    for i = 0 to 2 do
      let v = d.(i) in
        draw ((map_vect (fun pt -> (xx v pt + orx, yy v pt + ory))
            f.(i)),
          couleur_binaire_de_couleur GRIS)
    done
;;

(* affichage du cube en grisé *)
let tout_griser orx ory =
  do_list (griser orx ory) indices
;;

(* affichage du cube dans l'état mv *)
let affiche_mouv orx ory mv =
  do_list (fun x -> affiche orx ory (fun_of_mv1 mv x) x) indices
;;

(* événements souris -----------------------------------*)

(* est_dans face renvoie 'true' ssi le point 'x' est intérieur au contour plan 'p'*)
let est_dans_face p x =
  let prefix /-/ v1 v2 = let v = make_vect 3 0 in
      for i = 0 to 1 do
        v.(i) <- v1.(i) - v2.(i)
      done;
      v
  and det v1 v2 = v1.(0) * v2.(1) - v1.(1) * v2.(0)
  in
    let p0 = p.(0) /-/ x
    and p1 = p.(1) /-/ x
    and p2 = p.(2) /-/ x
    and p3 = p.(3) /-/ x
    in
      det p0 p1 * det p1 p2 > 0
      && det p1 p2 * det p2 p3 > 0
      && det p2 p3 * det p3 p0 > 0
;;

(* resultat d'un clic dans le panneau de gauche *)
let resultat = ref (- 1, - 1);;


(* traitement d'un clic dans le panneau de droite *)
let traitement (ci, fi) =
  let c, d, _ = faces mvi.(ci).ct
  in
    (c, d.(fi))
;;

exception Fin_de_recherche;;(* on a cliqué sur une face de minicube dans le panneau gauche *)
exception Fin_de_recherche1;;(* on a cliqué sur une face de minicube dans le panneau droit *)

(* gestion souris du panneau gauche *)
let gestion_a_gauche mousex mousey =
  let i1 = ref (- 1) and i2 = ref (- 1) in
    let action ct mousex mousey =
      let c, d, f = faces ct in
        for j = 0 to 2 do
          let face = map_vect (fun pt -> [|xx d.(j) pt; yy d.(j) pt|]) f.(j) in
            if est_dans_face face [|mousex; mousey|] then (i1 := j; raise Fin_de_recherche)
        done
    in
      try
        for k = 0 to vect_length mvi - 1 do
          i2 := k;
          action mvi.(k).ct mousex mousey
        done;
      with Fin_de_recherche ->
            if mvi.(!i2).videG then resultat := (- 1, - 1) else resultat := (!i2, !i1)
;;

(* gestion souris du panneau droit *)
let gestion_a_droite mousex mousey =
  let i1 = ref (- 1) and i2 = ref (- 1) in
    let action ct mousex mousey =
      let c, d, f = faces ct in
        for j = 0 to 2 do
          let face = map_vect (fun pt -> [|xx d.(j) pt + (largeur / 2); yy d.(j) pt|]) f.(j) in
            if est_dans_face face [|mousex; mousey|] then (
                i1 := j; raise Fin_de_recherche1
              )
        done
    in
      try
        for k = 0 to vect_length mvi - 1 do
          i2 := k;
          action mvi.(k).ct mousex mousey
        done;
      with Fin_de_recherche1 ->
            let (ci, fi) = !resultat in
              if mvi.(!i2).videD then
                (
                  let (ci, fi) = !resultat in
                    if ci <> - 1 then
                      let (v1, v2) = traitement (ci, fi)
                      and (w1, w2) = traitement (!i2, !i1) in
                        let mat = hd (select (fun m -> v1 /:/ m = w1 && v2 /:/ m = w2) groupe_du_cube)
                        in
                          mvi.(!i2).mat <- transpose mat;
                          mvi.(!i2).videD <- false;
                          mvi.(ci).videG <- true;
                          affiche (largeur / 2) 0 mat mvi.(ci).ct;
                          griser 0 0 mvi.(ci).ct;
                          resultat := (- 1, - 1)
                )
              else
                (
                  let (w1, w2) = traitement (!i2, !i1) in
                    try
                      let mat = mvi.(!i2).mat in
                        let i3 = ref (- 1) in
                          try
                            for k = 0 to vect_length mvi - 1 do
                              i3 := k;
                              if mvi.(k).ct = w1 /:/ mat then raise Fin_de_recherche1
                            done
                          with Fin_de_recherche1 ->
                                (
                                  mvi.(!i3).videG <- false;
                                  mvi.(!i2).videD <- true;
                                  mvi.(!i2).mat <- matrice_nulle;
                                  affiche 0 0 id mvi.(!i3).ct;
                                  griser (largeur / 2) 0 mvi.(!i2).ct;
                                  resultat := (- 1, - 1)
                                )
                    with Not_found -> (print_string "Not Found"; print_newline ())
                )
;;


(*-------------------------------------------------------------------------------------------*)
(* TEST DE VALIDITÉ *)
(*-------------------------------------------------------------------------------------------*)

let marque x = [|0; 0; x.(2)|];;


(* morphisme 's: M -> S' et section 'l: S -> M' *)
(* construction d'une section 'l' de la suite exacte '0 -> K -> M -> S -> 0' *)
(* En Caml on représente la sujection 's' par 'sur', la section 'l' par 'sec' et 'gij' par 'gg i j' *)

(* éléments g_{ij} alias gg i j de G servant à construire cette section *)
let gg i j =
  try
    let critere i j g = i /:/ g = j && marque i /:/ g = marque j
    in
      hd (select (critere i j) groupe_du_cube)
  with Failure "hd" -> id
        (* cette liste devrait toujours contenir exactement un élément *)
;;

(* stabilisateur du coin d'indice i *)
let st i =
  let k = [|1; 1; 1|] and m = [|[|0; 0; 1|]; [|1; 0; 0|]; [|0; 1; 0|]|] (* h_kappa pour C *)
  in
    gg i k /./ m /./ gg k i
;;

(* décomposition 'm = ker m /*/ sec (sur m)' d'un mouvement 'm' *)
(* avec 'ker m' élément du noyau de 'sur' *)
(* 'p' pour 'permutation': 'p = sur m' est la permutation 'p' des indices associée au mouvement 'm' *)
let sec p = mv1_of_fun (fun i -> gg i (p i));;
let sur m = fun i -> i /:/ fun_of_mv1 m i;;
let ker m = m /*/ inverse (sec (sur m));;

(* rotation totale des coins *)
let rtc m =
  let rtc_aux k = let f = fun_of_mv1 k in
      let indexc i = if f i = st i then 1
        else if f i = transpose (st i) then 2 else 0 in
        (list_it (prefix +) (map indexc indices) 0) mod 3
  in rtc_aux (ker m)
;;

(* test d'appartenance d'un mouvement au sous-groupe R *)
let est_dans_R m = rtc m = 0;;


(*-------------------------------------------------------------------------------------------*)
(* BOUCLE DE CONFIGURATION DU CUBE *)
(*-------------------------------------------------------------------------------------------*)

exception Quitter;;

let initialise () =
  for i = 0 to list_length indices - 1 do
    let v = vect_of_list indices in
      mvi.(i) <- {videG = false; videD = true; ct = v.(i); mat = matrice_nulle}
  done;
  resultat := (- 1, - 1);
;;

let boucle_saisie s =
  initialise ();
  graphics__open_graph s;
  graphics__set_window_title "Composer un cube";
  affiche_mouv 0 0 e; (* affichage à gauche du cube non mélangé *)
  tout_griser (largeur / 2) 0; (* affichage à droite du cube grisé *)
  graphics__set_color graphics__black;
  graphics__moveto (largeur / 2) 0;
  graphics__lineto (largeur / 2) hauteur; (* cloison entre les deux panneaux *)
  printf__printf "\n- Pour transférer un minicube de gauche à droite, cliquer sur une de ses faces puis cliquer à droite sur la destination de cette face (une face grise).\n";
  printf__printf "- Pour ramener un minicube de droite à sa place à gauche, cliquer sur une de ses faces.\n\n";
  printf__printf "- La rotation totale des minicubes est vérifiée quand tous les minicubes sont à droite.\n";
  printf__printf "- En cas de copie d'un cube correct, une rotation totale non nulle implique une erreur dans cette copie.\n\n";
  printf__printf "- Si la rotation totale n'est pas nulle quand on clique sur 'Quitter', la copie est remplacée par une configuration correcte par rotation du minicube frontal en haut à droite.\n";
  printf__printf "- Mais cette configuration correcte n'est généralement pas celle qu'on a voulu copier!\n";
  print_newline ();
  try
    let largeur1, hauteur1 = graphics__text_size "Quitter"
    in
      let bouton_quitter = bouton "Quitter" (graphics__size_x () - largeur1 - 20) (graphics__size_y () - hauteur1 - 20)
        (largeur1 + 10) (hauteur1 + 10) JAUNE (fun () -> raise Quitter)
      in
        dessine_bouton bouton_quitter;
        while true do
          let status = graphics__wait_next_event [graphics__Button_down] in
            let mousex = status.graphics__mouse_x and mousey = status.graphics__mouse_y
            in
              if status.graphics__button then (
                  gestion_bouton bouton_quitter mousex mousey;
                  gestion_a_gauche mousex mousey;
                  gestion_a_droite mousex mousey;
                );
              if for_all (fun t -> mvi.(t).videG) (liste (vect_length mvi)) then (
                  mouvement.mv1 <- inverse (map (fun i -> let elem = mvi.(i) in (elem.ct, elem.mat)) (liste (vect_length mvi)));
                  let rt = rtc mouvement.mv1 in
                    printf__printf "rotation totale : %d" rt;
                    print_newline ();
                );
        done
  with Quitter ->
        if for_all (fun t -> mvi.(t).videG) (liste (vect_length mvi)) then (
            mouvement.mv1 <- inverse (map (fun i -> let elem = mvi.(i) in (elem.ct, elem.mat)) (liste (vect_length mvi)));
            let rt = rtc mouvement.mv1 in
              printf__printf "rotation totale : %d" rt;
              print_newline ();
              if rt <> 0 then (
                  mouvement.mv1 <- mouvement.mv1 /*/ mv1_of_fun (fun t -> if t <> [|1; 1; 1|] then id else if rt = 1 then transpose (st [|1; 1; 1|]) else st [|1; 1; 1|]);
                  affiche_mouv (largeur / 2) 0 mouvement.mv1;
                  printf__printf "mouvement corrigé : rotation totale : %d" (rtc mouvement.mv1);
                  print_newline ()
                );
          );
        graphics__close_graph ();
;;

let l = string_of_int largeur and h = string_of_int hauteur in
  boucle_saisie (" " ^ l ^ "x" ^ h);;


(*-------------------------------------------------------------------------------------------*)
(* RÉSOLUTION DU CUBE *)
(*-------------------------------------------------------------------------------------------*)

(* liste des coins : ici il n'y a que des coins! *)
let coins = indices;;

(* résolution du cube 2x2 *)
exception Orienter_les_coins;;
exception Descendre_coin;;
exception Remonter_coin;;
exception Placer_les_coins;;

let dessine_cube context mv1 = affiche_mouv 0 0 mv1;;

let anterieure = [|1;0;0|];;
let droite = [|0;1;0|];;
let haute = [|0;0;1|];;
let posterieure = [|-1;0;0|];;
let gauche = [|0;-1;0|];;
let basse = [|0;0;-1|];;

let code_face face =
match face with
    | ANTERIEURE -> anterieure
    | DROITE -> droite
    | HAUTE -> haute
    | POSTERIEURE -> posterieure
    | GAUCHE -> gauche
    | BASSE -> basse
;;
let face_code code =
  let l_faces = [ANTERIEURE; DROITE; HAUTE; POSTERIEURE; GAUCHE; BASSE]
  in let l_codes = map code_face l_faces
    in try assoc code (map2 (fun x y -> (x, y)) l_codes l_faces)
      with not_found -> failwith "face_code"
;;

let nom_de_face v =
  match face_code v with
    | POSTERIEURE -> "p"
    | ANTERIEURE -> "a"
    | GAUCHE -> "g"
    | DROITE -> "d"
    | BASSE -> "b"
    | HAUTE -> "h"
;;


let resoudre_le_mouvement mouvement en_dessinant en_imprimant =
  (* (y, m) = pos0 x : le minicube d'indice x est à l'emplacement d'indice y et m est sa matrice *)
  (* de déplacement (telle que y=xm) (repère ADH) *)
  let context = {matrice = id} in
    let liste_ops = ref []
    and dessine () = if en_dessinant then dessine_cube context mouvement.mv1
    in
      let rotations_faces liste_ops =
        let fct x () =
          let t = x /:/ transpose context.matrice in
            mouvement.mv1 <- mouvement.mv1 /*/ rub t;
            if en_imprimant then print_string (nom_de_face t ^ " ");
            liste_ops := (nom_de_face t) :: !liste_ops;
            dessine ()
        and fct' x () =
          let t = x /:/ transpose context.matrice in
            mouvement.mv1 <- mouvement.mv1 /*/ rub' t;
            if en_imprimant then print_string (nom_de_face t ^ "' ");
            liste_ops := (nom_de_face t ^ "'") :: !liste_ops;
            dessine ()
        in
          let (a, d, h) = vect (map_vect fct id)
          and (a', d', h') = vect (map_vect fct' id)
          and (p, g, b) = vect (map_vect fct idm)
          and (p', g', b') = vect (map_vect fct' idm)
          in (OPS (a, d, h), OPS (a', d', h'), OPS (p, g, b), OPS (p', g', b'))
      and rotations_cube () =
        let rotation pp () = context.matrice <- context.matrice /./ pp;
          dessine () in
          let (a, d, h) = vect (map_vect rotation (map_vect rot id))
          and (a', d', h') = vect (map_vect rotation (map_vect rot' id))
          in
            (OPS (a, d, h), OPS (a', d', h'))
      in
        
        let pos0 =
          let pos mv1 x =
            let p = context.matrice in
              let m = transpose p /./ (fun_of_mv1 mv1) (x /:/ transpose p) /./ p
              in
                x /:/ m, m
          in
            (fun x -> pos mouvement.mv1 x)
        and
        (OPS (a, d, h), OPS (a', d', h'), OPS (p, g, b), OPS (p', g', b')) = rotations_faces liste_ops
        and
        (OPS (_, _, h0), OPS (_, _, h0')) = rotations_cube ()
        in
          
          
          let niveau_superieur () =
            
            (* niveau supérieur *)
            
            let placer_et_orienter_les_coins () =
              let descendre_coin () =
                let w, m = pos0 [|1; 1; 1|] in
                  if (w <> [|1; 1; 1|]) || (m <> id) then
                    match vect w with
                      | (- 1, 1, 1) -> exe [p'; b'; p]
                      | (- 1, - 1, 1) -> exe [p; b; b; p']
                      | (1, - 1, 1) -> exe [g; b; g']
                      | (1, 1, 1) -> exe [a; b; a'; b']
                      | (- 1, 1, - 1) -> exe [b']
                      | (- 1, - 1, - 1) -> exe [b; b]
                      | (1, - 1, - 1) -> exe [b]
                      | (1, 1, - 1) -> ()
                      | _ -> raise Descendre_coin
              and remonter_coin () =
                let (v, m) = pos0 [|1; 1; 1|] in
                  if (v = [|1; 1; 1|]) && (m = id) then ()
                  else
                    let w = m.(2) in match vect w with
                        | (1, 0, 0) -> exe [d; a'; d'; a]
                        | (0, 1, 0) -> exe [a'; d; a; d']
                        | (0, 0, - 1) -> exe [a; b'; a'; b; b; d; a'; d'; a]
                        | _ -> raise Remonter_coin
              in
                for i = 0 to 3 do
                  descendre_coin ();
                  remonter_coin ();
                  exe [h0]
                done;
            
            in
              placer_et_orienter_les_coins ()
          
          
          and niveau_inferieur () =
            
            (* niveau inférieur *)
            
            let placer_les_coins () =
              let permuter () =
                (* laisse fixe le coin frontal droit et permute circulairement 
          les autres dans le sens direct vu d'en bas *)
                exe [b; a; b'; p'; b; a'; b'; p]
              and permuter' () =
                (* laisse fixe le coin frontal droit et permute circulairement 
          les autres dans le sens indirect vu d'en bas *)
                exe [p'; b; a; b'; p; b; a'; b']
              in
                let chercher_un_coin_bien_place () =
                  let i = ref 0 in
                    while !i < 4 && fst (pos0 [|1; 1; - 1|]) <> [|1; 1; - 1|] do
                      exe [h0];
                      incr i
                    done;
                    !i
                in
                  let j = chercher_un_coin_bien_place () in
                    
                    if j = 4 (* aucun coin bien placé *) then (
                        permuter ();
                        let _ = chercher_un_coin_bien_place () in ()
                      )
                    else ();
                    
                    let v, _ = pos0 [|- 1; - 1; - 1|] in match vect v with
                        | (1, - 1, - 1) -> permuter ()
                        | (- 1, 1, - 1) -> permuter' ()
                        | (- 1, - 1, - 1) -> ()
                        | _ -> raise Placer_les_coins
            
            
            and orienter_les_coins () =
              let faire_tourner () =
                (* fait tourner les coins frontaux inférieurs sur eux-mêmes: 
          le coin gauche dans le sens direct, le coin droit en sens inverse *)
                exe [p'; b'; p; b'; p'; b; b; p];
                exe [a; b; a'; b; a; b; b; a']
              and
              faire_tourner' () =
                (* fait tourner les coins frontaux inférieurs sur eux-mêmes:
          le coin droit dans le sens direct, le coin gauche en sens inverse *)
                exe [a; b; b; a'; b'; a; b'; a'];
                exe [p'; b'; b'; p; b; p'; b; p]
              in
                let orienter_frontal_inferieur_droit () =
                  let _, m = pos0 [|1; 1; - 1|] in
                    let v = m.(2) in
                      match vect v with
                        | (0, 0, 1) -> ()
                        | (- 1, 0, 0) -> faire_tourner' ()
                        | (0, - 1, 0) -> faire_tourner ()
                        | _ -> raise Orienter_les_coins
                in
                  for i = 0 to 2 do
                    orienter_frontal_inferieur_droit ();
                    exe [h0']
                  done
            
            in
            (* faire un quart de tour si la permutation des coins est impaire *)
            (* car 'placer les coins' suppose cette permutation paire *)
              if (sign coins (sur mouvement.mv1) = - 1) then exe [b]
              else ();
              placer_les_coins ();
              orienter_les_coins ()
          
          in
            try
              niveau_superieur ();
              niveau_inferieur ();
              context.matrice <- id;
              dessine ();
              rev !liste_ops
            with
              | Orienter_les_coins ->
                  print_string "erreur dans orienter_les_coins\n"; rev !liste_ops
              | Placer_les_coins ->
                  print_string "erreur dans placer_les_coins\n"; rev !liste_ops
;;


(*-------------------------------------------------------------------------------------------*)
(* BOUTONS POUR LA MANIPULATION DU CUBE *)
(*-------------------------------------------------------------------------------------------*)

let op_rub mouv context = let p = transpose context.matrice in
    map (fun t -> (fun () -> mouv.mv1 <- mouv.mv1 /*/ rub (t /:/ p); affiche_mouv 0 0 mouv.mv1)) ((list_of_vect id) @ (list_of_vect idm))
    @ map (fun t -> (fun () -> mouv.mv1 <- mouv.mv1 /*/ rub' (t /:/ p); affiche_mouv 0 0 mouv.mv1)) ((list_of_vect id) @ (list_of_vect idm))
;;

let cree_boutons mouvement context =
  let titres =
    [|
      "a"; "a'"; "h"; "h'"; "d"; "d'";
      "p"; "p'"; "b"; "b'"; "g"; "g'"
    |]
  in
    let set_actions boutons =
      let v = vect_of_list (op_rub mouvement context) in
        
        let a = v.(0) and d = v.(1) and h = v.(2) and p = v.(3) and g = v.(4) and b = v.(5)
        and a' = v.(6) and d' = v.(7) and h' = v.(8) and p' = v.(9) and g' = v.(10) and b' = v.(11)
        in let w = [|a; a'; h; h'; d; d'; p; p'; b; b'; g; g'|]
          in for i = 0 to vect_length v - 1 do
              set_action boutons.(i) w.(i)
            done
    in let boutons = make_vect 12 (bouton "" 0 0 0 0 BLANC (fun () -> ()))
      in
        for i = 0 to 11 do
          boutons.(i) <- {titre = titres.(i); orx = i * 34 * 3 / 2; ory = 0; hauteur = 30 * 3 / 2;
            largeur = 34 * 3 / 2; couleur = BLANC; action = fun () -> ()}
        done;
        set_actions boutons;
        boutons
;;

let dessine_boutons mouvement context =
  let couleur_titre titre =
    let face titre = match titre with
        | "A" | "A'" | "a" | "a'" -> [|1; 0; 0|]
        | "D" | "D'" | "d" | "d'" -> [|0; 1; 0|]
        | "H" | "H'" | "h" | "h'" -> [|0; 0; 1|]
        | "p" | "p'" -> [|- 1; 0; 0|]
        | "g" | "g'" -> [|0; - 1; 0|]
        | "b" | "b'" -> [|0; 0; - 1|]
        | _ -> failwith "face"
    in
      couleur_de ((face titre) /:/ transpose context.matrice)
  in
    let boutons = cree_boutons mouvement context
    in
      for i = 0 to vect_length boutons - 1 do
        boutons.(i).couleur <- couleur_titre boutons.(i).titre;
        dessine_bouton boutons.(i)
      done
;;

let gestion_boutons boutons mouvement context mousex mousey =
  for i = 0 to vect_length boutons - 1 do
    gestion_bouton boutons.(i) mousex mousey
  done;
  dessine_boutons mouvement context
;;


(*-------------------------------------------------------------------------------------------*)
(* BOUCLE DE MANIPULATION DU CUBE *)
(*-------------------------------------------------------------------------------------------*)

let boucle1 mouvement context action1 action2 =
  graphics__open_graph " 612x612";
  graphics__set_window_title "Résoudre le cube";
  ux := 20;
  uy := 20;
  ox := 306;
  oy := 320;
  try
    affiche_mouv 0 0 mouvement.mv1;
    let largeur, hauteur = graphics__text_size "Résoudre"
    and largeur1, hauteur1 = graphics__text_size "Quitter"
    and largeur2, hauteur2 = graphics__text_size "Restaurer"
    and boutons = cree_boutons mouvement context;
    
    in
      let
      bouton_quitter = bouton "Quitter" (graphics__size_x () - largeur1 - 20) (graphics__size_y () - hauteur1 - 20)
        (largeur1 + 10) (hauteur1 + 10) JAUNE (fun () -> raise Quitter)
      and bouton_resoudre = bouton "Résoudre" 20 (graphics__size_y () - hauteur - 20)
        (largeur + 10) (hauteur + 10) JAUNE action1
      and bouton_restaurer = bouton "Restaurer" ((graphics__size_x () - largeur2) / 2 - 10) (graphics__size_y () - hauteur2 - 20)
        (largeur2 + 10) (hauteur2 + 10) JAUNE action2
      in
        dessine_bouton bouton_resoudre;
        dessine_bouton bouton_quitter;
        dessine_bouton bouton_restaurer;
        dessine_boutons mouvement context;
        while true do
          let status = graphics__wait_next_event [graphics__Button_down] in
            let mousex = status.graphics__mouse_x and mousey = status.graphics__mouse_y
            in
              if status.graphics__button then (
                  gestion_bouton bouton_resoudre mousex mousey;
                  gestion_bouton bouton_quitter mousex mousey;
                  gestion_bouton bouton_restaurer mousex mousey;
                  gestion_boutons boutons mouvement context mousex mousey;
                )
        done;
  with Quitter -> graphics__close_graph ()
;;

let boucle mouvement =
  let mouvement_bis = {mv1 = mouvement.mv1} in
    boucle1 mouvement_bis {matrice = id}
    (fun () -> let _ = resoudre_le_mouvement mouvement_bis true true in (); print_newline ())
    (fun () -> mouvement_bis.mv1 <- mouvement.mv1; affiche_mouv 0 0 mouvement_bis.mv1);;

boucle mouvement;;

 
